@@ -8,7 +8,7 @@ use crate::event::Event;
 use crate::query::Query;
 use crate::request::Wrapper;
 use crate::{
-    response, Client, Error, Id, Request, Response, Result, Scheme, SimpleRequest, Subscription,
+    error::Error, response, Client, Id, Request, Response, Scheme, SimpleRequest, Subscription,
     SubscriptionClient, Url,
 };
 use async_trait::async_trait;
@@ -43,6 +43,9 @@ const RECV_TIMEOUT: Duration = Duration::from_secs(RECV_TIMEOUT_SECONDS);
 // Taken from https://github.com/tendermint/tendermint/blob/309e29c245a01825fc9630103311fd04de99fa5e/rpc/jsonrpc/server/ws_handler.go#L28
 const PING_INTERVAL: Duration = Duration::from_secs((RECV_TIMEOUT_SECONDS * 9) / 10);
 
+/// Low-level WebSocket configuration
+pub use async_tungstenite::tungstenite::protocol::WebSocketConfig;
+
 /// Tendermint RPC client that provides access to all RPC functionality
 /// (including [`Event`] subscription) over a WebSocket connection.
 ///
@@ -132,23 +135,39 @@ impl WebSocketClient {
     /// Tendermint node's RPC endpoint.
     ///
     /// Supports both `ws://` and `wss://` protocols.
-    pub async fn new<U>(url: U) -> Result<(Self, WebSocketClientDriver)>
+    pub async fn new<U>(url: U) -> Result<(Self, WebSocketClientDriver), Error>
+    where
+        U: TryInto<WebSocketClientUrl, Error = Error>,
+    {
+        Self::new_with_config(url, None).await
+    }
+
+    /// Construct a new WebSocket-based client connecting to the given
+    /// Tendermint node's RPC endpoint.
+    ///
+    /// Supports both `ws://` and `wss://` protocols.
+    pub async fn new_with_config<U>(
+        url: U,
+        config: Option<WebSocketConfig>,
+    ) -> Result<(Self, WebSocketClientDriver), Error>
     where
         U: TryInto<WebSocketClientUrl, Error = Error>,
     {
         let url = url.try_into()?;
+
         let (inner, driver) = if url.0.is_secure() {
-            sealed::WebSocketClient::new_secure(url.0).await?
+            sealed::WebSocketClient::new_secure(url.0, config).await?
         } else {
-            sealed::WebSocketClient::new_unsecure(url.0).await?
+            sealed::WebSocketClient::new_unsecure(url.0, config).await?
         };
+
         Ok((Self { inner }, driver))
     }
 }
 
 #[async_trait]
 impl Client for WebSocketClient {
-    async fn perform<R>(&self, request: R) -> Result<<R as Request>::Response>
+    async fn perform<R>(&self, request: R) -> Result<<R as Request>::Response, Error>
     where
         R: SimpleRequest,
     {
@@ -158,15 +177,15 @@ impl Client for WebSocketClient {
 
 #[async_trait]
 impl SubscriptionClient for WebSocketClient {
-    async fn subscribe(&self, query: Query) -> Result<Subscription> {
+    async fn subscribe(&self, query: Query) -> Result<Subscription, Error> {
         self.inner.subscribe(query).await
     }
 
-    async fn unsubscribe(&self, query: Query) -> Result<()> {
+    async fn unsubscribe(&self, query: Query) -> Result<(), Error> {
         self.inner.unsubscribe(query).await
     }
 
-    fn close(self) -> Result<()> {
+    fn close(self) -> Result<(), Error> {
         self.inner.close()
     }
 }
@@ -180,10 +199,10 @@ pub struct WebSocketClientUrl(Url);
 impl TryFrom<Url> for WebSocketClientUrl {
     type Error = Error;
 
-    fn try_from(value: Url) -> Result<Self> {
+    fn try_from(value: Url) -> Result<Self, Error> {
         match value.scheme() {
             Scheme::WebSocket | Scheme::SecureWebSocket => Ok(Self(value)),
-            _ => Err(Error::invalid_params(&format!(
+            _ => Err(Error::invalid_params(format!(
                 "cannot use URL {} with WebSocket clients",
                 value
             ))),
@@ -194,7 +213,7 @@ impl TryFrom<Url> for WebSocketClientUrl {
 impl FromStr for WebSocketClientUrl {
     type Err = Error;
 
-    fn from_str(s: &str) -> Result<Self> {
+    fn from_str(s: &str) -> Result<Self, Error> {
         let url: Url = s.parse()?;
         url.try_into()
     }
@@ -203,7 +222,7 @@ impl FromStr for WebSocketClientUrl {
 impl TryFrom<&str> for WebSocketClientUrl {
     type Error = Error;
 
-    fn try_from(value: &str) -> Result<Self> {
+    fn try_from(value: &str) -> Result<Self, Error> {
         value.parse()
     }
 }
@@ -211,7 +230,7 @@ impl TryFrom<&str> for WebSocketClientUrl {
 impl TryFrom<net::Address> for WebSocketClientUrl {
     type Error = Error;
 
-    fn try_from(value: net::Address) -> Result<Self> {
+    fn try_from(value: net::Address) -> Result<Self, Error> {
         match value {
             net::Address::Tcp {
                 peer_id: _,
@@ -219,7 +238,7 @@ impl TryFrom<net::Address> for WebSocketClientUrl {
                 port,
             } => format!("ws://{}:{}/websocket", host, port).parse(),
             net::Address::Unix { .. } => Err(Error::invalid_params(
-                "only TCP-based node addresses are supported",
+                "only TCP-based node addresses are supported".to_string(),
             )),
         }
     }
@@ -236,12 +255,18 @@ mod sealed {
         DriverCommand, SimpleRequestCommand, SubscribeCommand, UnsubscribeCommand,
         WebSocketClientDriver,
     };
+
     use crate::client::sync::{unbounded, ChannelTx};
     use crate::query::Query;
     use crate::request::Wrapper;
     use crate::utils::uuid_str;
-    use crate::{Error, Response, Result, SimpleRequest, Subscription, Url};
-    use async_tungstenite::tokio::{connect_async, connect_async_with_tls_connector};
+    use crate::{Error, Response, SimpleRequest, Subscription, Url};
+
+    use async_tungstenite::{
+        tokio::{connect_async_with_config, connect_async_with_tls_connector_and_config},
+        tungstenite::protocol::WebSocketConfig,
+    };
+
     use tracing::debug;
 
     /// Marker for the [`AsyncTungsteniteClient`] for clients operating over
@@ -275,19 +300,25 @@ mod sealed {
         /// this driver becomes the responsibility of the client owner, and must be
         /// executed in a separate asynchronous context to the client to ensure it
         /// doesn't block the client.
-        pub async fn new(url: Url) -> Result<(Self, WebSocketClientDriver)> {
+        pub async fn new(
+            url: Url,
+            config: Option<WebSocketConfig>,
+        ) -> Result<(Self, WebSocketClientDriver), Error> {
             let url = url.to_string();
             debug!("Connecting to unsecure WebSocket endpoint: {}", url);
-            let (stream, _response) = connect_async(url).await?;
+
+            let (stream, _response) = connect_async_with_config(url, config)
+                .await
+                .map_err(Error::tungstenite)?;
+
             let (cmd_tx, cmd_rx) = unbounded();
             let driver = WebSocketClientDriver::new(stream, cmd_rx);
-            Ok((
-                Self {
-                    cmd_tx,
-                    _client_type: Default::default(),
-                },
-                driver,
-            ))
+            let client = Self {
+                cmd_tx,
+                _client_type: Default::default(),
+            };
+
+            Ok((client, driver))
         }
     }
 
@@ -301,35 +332,37 @@ mod sealed {
         /// this driver becomes the responsibility of the client owner, and must be
         /// executed in a separate asynchronous context to the client to ensure it
         /// doesn't block the client.
-        pub async fn new(url: Url) -> Result<(Self, WebSocketClientDriver)> {
+        pub async fn new(
+            url: Url,
+            config: Option<WebSocketConfig>,
+        ) -> Result<(Self, WebSocketClientDriver), Error> {
             let url = url.to_string();
             debug!("Connecting to secure WebSocket endpoint: {}", url);
+
             // Not supplying a connector means async_tungstenite will create the
             // connector for us.
-            let (stream, _response) = connect_async_with_tls_connector(url, None).await?;
+            let (stream, _response) =
+                connect_async_with_tls_connector_and_config(url, None, config)
+                    .await
+                    .map_err(Error::tungstenite)?;
+
             let (cmd_tx, cmd_rx) = unbounded();
             let driver = WebSocketClientDriver::new(stream, cmd_rx);
-            Ok((
-                Self {
-                    cmd_tx,
-                    _client_type: Default::default(),
-                },
-                driver,
-            ))
+            let client = Self {
+                cmd_tx,
+                _client_type: Default::default(),
+            };
+
+            Ok((client, driver))
         }
     }
 
     impl<C> AsyncTungsteniteClient<C> {
-        fn send_cmd(&self, cmd: DriverCommand) -> Result<()> {
-            self.cmd_tx.send(cmd).map_err(|e| {
-                Error::client_internal_error(format!(
-                    "failed to send command to client driver: {}",
-                    e
-                ))
-            })
+        fn send_cmd(&self, cmd: DriverCommand) -> Result<(), Error> {
+            self.cmd_tx.send(cmd)
         }
 
-        pub async fn perform<R>(&self, request: R) -> Result<R::Response>
+        pub async fn perform<R>(&self, request: R) -> Result<R::Response, Error>
         where
             R: SimpleRequest,
         {
@@ -343,15 +376,13 @@ mod sealed {
                 response_tx,
             }))?;
             let response = response_rx.recv().await.ok_or_else(|| {
-                Error::client_internal_error(
-                    "failed to hear back from WebSocket driver".to_string(),
-                )
+                Error::client_internal("failed to hear back from WebSocket driver".to_string())
             })??;
             tracing::debug!("Incoming response: {}", response);
             R::Response::from_string(response)
         }
 
-        pub async fn subscribe(&self, query: Query) -> Result<Subscription> {
+        pub async fn subscribe(&self, query: Query) -> Result<Subscription, Error> {
             let (subscription_tx, subscription_rx) = unbounded();
             let (response_tx, mut response_rx) = unbounded();
             // By default we use UUIDs to differentiate subscriptions
@@ -364,29 +395,25 @@ mod sealed {
             }))?;
             // Make sure our subscription request went through successfully.
             let _ = response_rx.recv().await.ok_or_else(|| {
-                Error::client_internal_error(
-                    "failed to hear back from WebSocket driver".to_string(),
-                )
+                Error::client_internal("failed to hear back from WebSocket driver".to_string())
             })??;
             Ok(Subscription::new(id, query, subscription_rx))
         }
 
-        pub async fn unsubscribe(&self, query: Query) -> Result<()> {
+        pub async fn unsubscribe(&self, query: Query) -> Result<(), Error> {
             let (response_tx, mut response_rx) = unbounded();
             self.send_cmd(DriverCommand::Unsubscribe(UnsubscribeCommand {
                 query: query.to_string(),
                 response_tx,
             }))?;
             let _ = response_rx.recv().await.ok_or_else(|| {
-                Error::client_internal_error(
-                    "failed to hear back from WebSocket driver".to_string(),
-                )
+                Error::client_internal("failed to hear back from WebSocket driver".to_string())
             })??;
             Ok(())
         }
 
         /// Signals to the driver that it must terminate.
-        pub fn close(self) -> Result<()> {
+        pub fn close(self) -> Result<(), Error> {
             self.send_cmd(DriverCommand::Terminate)
         }
     }
@@ -400,17 +427,23 @@ mod sealed {
     }
 
     impl WebSocketClient {
-        pub async fn new_unsecure(url: Url) -> Result<(Self, WebSocketClientDriver)> {
-            let (client, driver) = AsyncTungsteniteClient::<Unsecure>::new(url).await?;
+        pub async fn new_unsecure(
+            url: Url,
+            config: Option<WebSocketConfig>,
+        ) -> Result<(Self, WebSocketClientDriver), Error> {
+            let (client, driver) = AsyncTungsteniteClient::<Unsecure>::new(url, config).await?;
             Ok((Self::Unsecure(client), driver))
         }
 
-        pub async fn new_secure(url: Url) -> Result<(Self, WebSocketClientDriver)> {
-            let (client, driver) = AsyncTungsteniteClient::<Secure>::new(url).await?;
+        pub async fn new_secure(
+            url: Url,
+            config: Option<WebSocketConfig>,
+        ) -> Result<(Self, WebSocketClientDriver), Error> {
+            let (client, driver) = AsyncTungsteniteClient::<Secure>::new(url, config).await?;
             Ok((Self::Secure(client), driver))
         }
 
-        pub async fn perform<R>(&self, request: R) -> Result<R::Response>
+        pub async fn perform<R>(&self, request: R) -> Result<R::Response, Error>
         where
             R: SimpleRequest,
         {
@@ -420,21 +453,21 @@ mod sealed {
             }
         }
 
-        pub async fn subscribe(&self, query: Query) -> Result<Subscription> {
+        pub async fn subscribe(&self, query: Query) -> Result<Subscription, Error> {
             match self {
                 WebSocketClient::Unsecure(c) => c.subscribe(query).await,
                 WebSocketClient::Secure(c) => c.subscribe(query).await,
             }
         }
 
-        pub async fn unsubscribe(&self, query: Query) -> Result<()> {
+        pub async fn unsubscribe(&self, query: Query) -> Result<(), Error> {
             match self {
                 WebSocketClient::Unsecure(c) => c.unsubscribe(query).await,
                 WebSocketClient::Secure(c) => c.unsubscribe(query).await,
             }
         }
 
-        pub fn close(self) -> Result<()> {
+        pub fn close(self) -> Result<(), Error> {
             match self {
                 WebSocketClient::Unsecure(c) => c.close(),
                 WebSocketClient::Secure(c) => c.close(),
@@ -465,7 +498,7 @@ struct SubscribeCommand {
     // Where to send subscription events.
     subscription_tx: SubscriptionTx,
     // Where to send the result of the subscription request.
-    response_tx: ChannelTx<Result<()>>,
+    response_tx: ChannelTx<Result<(), Error>>,
 }
 
 #[derive(Debug, Clone)]
@@ -473,7 +506,7 @@ struct UnsubscribeCommand {
     // The query from which to unsubscribe.
     query: String,
     // Where to send the result of the unsubscribe request.
-    response_tx: ChannelTx<Result<()>>,
+    response_tx: ChannelTx<Result<(), Error>>,
 }
 
 #[derive(Debug, Clone)]
@@ -485,7 +518,7 @@ struct SimpleRequestCommand {
     // The wrapped and serialized JSON-RPC request.
     wrapped_request: String,
     // Where to send the result of the simple request.
-    response_tx: ChannelTx<Result<String>>,
+    response_tx: ChannelTx<Result<String, Error>>,
 }
 
 #[derive(Serialize, Deserialize, Debug, Clone)]
@@ -521,7 +554,7 @@ impl WebSocketClientDriver {
 
     /// Executes the WebSocket driver, which manages the underlying WebSocket
     /// transport.
-    pub async fn run(mut self) -> Result<()> {
+    pub async fn run(mut self) -> Result<(), Error> {
         let mut ping_interval =
             tokio::time::interval_at(Instant::now().add(PING_INTERVAL), PING_INTERVAL);
 
@@ -538,8 +571,9 @@ impl WebSocketClientDriver {
                         self.handle_incoming_msg(msg).await?
                     },
                     Err(e) => return Err(
-                        Error::websocket_error(
-                            format!("failed to read from WebSocket connection: {}", e),
+                        Error::web_socket(
+                            "failed to read from WebSocket connection".to_string(),
+                            e
                         ),
                     ),
                 },
@@ -551,22 +585,19 @@ impl WebSocketClientDriver {
                 },
                 _ = ping_interval.tick() => self.ping().await?,
                 _ = &mut recv_timeout => {
-                    return Err(Error::websocket_error(format!(
-                        "reading from WebSocket connection timed out after {} seconds",
-                        RECV_TIMEOUT.as_secs()
-                    )));
+                    return Err(Error::web_socket_timeout(RECV_TIMEOUT));
                 }
             }
         }
     }
 
-    async fn send_msg(&mut self, msg: Message) -> Result<()> {
+    async fn send_msg(&mut self, msg: Message) -> Result<(), Error> {
         self.stream.send(msg).await.map_err(|e| {
-            Error::websocket_error(format!("failed to write to WebSocket connection: {}", e))
+            Error::web_socket("failed to write to WebSocket connection".to_string(), e)
         })
     }
 
-    async fn send_request<R>(&mut self, wrapper: Wrapper<R>) -> Result<()>
+    async fn send_request<R>(&mut self, wrapper: Wrapper<R>) -> Result<(), Error>
     where
         R: Request,
     {
@@ -576,7 +607,7 @@ impl WebSocketClientDriver {
         .await
     }
 
-    async fn subscribe(&mut self, cmd: SubscribeCommand) -> Result<()> {
+    async fn subscribe(&mut self, cmd: SubscribeCommand) -> Result<(), Error> {
         // If we already have an active subscription for the given query,
         // there's no need to initiate another one. Just add this subscription
         // to the router.
@@ -601,7 +632,7 @@ impl WebSocketClientDriver {
         Ok(())
     }
 
-    async fn unsubscribe(&mut self, cmd: UnsubscribeCommand) -> Result<()> {
+    async fn unsubscribe(&mut self, cmd: UnsubscribeCommand) -> Result<(), Error> {
         // Terminate all subscriptions for this query immediately. This
         // prioritizes acknowledgement of the caller's wishes over networking
         // problems.
@@ -625,7 +656,7 @@ impl WebSocketClientDriver {
         Ok(())
     }
 
-    async fn simple_request(&mut self, cmd: SimpleRequestCommand) -> Result<()> {
+    async fn simple_request(&mut self, cmd: SimpleRequestCommand) -> Result<(), Error> {
         if let Err(e) = self
             .send_msg(Message::Text(cmd.wrapped_request.clone()))
             .await
@@ -638,7 +669,7 @@ impl WebSocketClientDriver {
         Ok(())
     }
 
-    async fn handle_incoming_msg(&mut self, msg: Message) -> Result<()> {
+    async fn handle_incoming_msg(&mut self, msg: Message) -> Result<(), Error> {
         match msg {
             Message::Text(s) => self.handle_text_msg(s).await,
             Message::Ping(v) => self.pong(v).await,
@@ -646,7 +677,7 @@ impl WebSocketClientDriver {
         }
     }
 
-    async fn handle_text_msg(&mut self, msg: String) -> Result<()> {
+    async fn handle_text_msg(&mut self, msg: String) -> Result<(), Error> {
         if let Ok(ev) = Event::from_string(&msg) {
             self.publish_event(ev).await;
             return Ok(());
@@ -726,7 +757,7 @@ impl WebSocketClientDriver {
         &mut self,
         pending_cmd: DriverCommand,
         response: String,
-    ) -> Result<()> {
+    ) -> Result<(), Error> {
         match pending_cmd {
             DriverCommand::Subscribe(cmd) => {
                 let (id, query, subscription_tx, response_tx) =
@@ -740,15 +771,15 @@ impl WebSocketClientDriver {
         }
     }
 
-    async fn pong(&mut self, v: Vec<u8>) -> Result<()> {
+    async fn pong(&mut self, v: Vec<u8>) -> Result<(), Error> {
         self.send_msg(Message::Pong(v)).await
     }
 
-    async fn ping(&mut self) -> Result<()> {
+    async fn ping(&mut self) -> Result<(), Error> {
         self.send_msg(Message::Ping(Vec::new())).await
     }
 
-    async fn close(mut self) -> Result<()> {
+    async fn close(mut self) -> Result<(), Error> {
         self.send_msg(Message::Close(Some(CloseFrame {
             code: CloseCode::Normal,
             reason: Cow::from("client closed WebSocket connection"),
@@ -783,8 +814,8 @@ mod test {
     // Interface to a driver that manages all incoming WebSocket connections.
     struct TestServer {
         node_addr: net::Address,
-        driver_hdl: JoinHandle<Result<()>>,
-        terminate_tx: ChannelTx<Result<()>>,
+        driver_hdl: JoinHandle<Result<(), Error>>,
+        terminate_tx: ChannelTx<Result<(), Error>>,
         event_tx: ChannelTx<Event>,
     }
 
@@ -809,11 +840,11 @@ mod test {
             }
         }
 
-        fn publish_event(&mut self, ev: Event) -> Result<()> {
+        fn publish_event(&mut self, ev: Event) -> Result<(), Error> {
             self.event_tx.send(ev)
         }
 
-        async fn terminate(self) -> Result<()> {
+        async fn terminate(self) -> Result<(), Error> {
             self.terminate_tx.send(Ok(())).unwrap();
             self.driver_hdl.await.unwrap()
         }
@@ -823,7 +854,7 @@ mod test {
     struct TestServerDriver {
         listener: TcpListener,
         event_rx: ChannelRx<Event>,
-        terminate_rx: ChannelRx<Result<()>>,
+        terminate_rx: ChannelRx<Result<(), Error>>,
         handlers: Vec<TestServerHandler>,
     }
 
@@ -831,7 +862,7 @@ mod test {
         fn new(
             listener: TcpListener,
             event_rx: ChannelRx<Event>,
-            terminate_rx: ChannelRx<Result<()>>,
+            terminate_rx: ChannelRx<Result<(), Error>>,
         ) -> Self {
             Self {
                 listener,
@@ -841,7 +872,7 @@ mod test {
             }
         }
 
-        async fn run(mut self) -> Result<()> {
+        async fn run(mut self) -> Result<(), Error> {
             loop {
                 tokio::select! {
                     Some(ev) = self.event_rx.recv() => self.publish_event(ev),
@@ -883,8 +914,8 @@ mod test {
     // Interface to a driver that manages a single incoming WebSocket
     // connection.
     struct TestServerHandler {
-        driver_hdl: JoinHandle<Result<()>>,
-        terminate_tx: ChannelTx<Result<()>>,
+        driver_hdl: JoinHandle<Result<(), Error>>,
+        terminate_tx: ChannelTx<Result<(), Error>>,
         event_tx: ChannelTx<Event>,
     }
 
@@ -907,7 +938,7 @@ mod test {
             let _ = self.event_tx.send(ev);
         }
 
-        async fn terminate(self) -> Result<()> {
+        async fn terminate(self) -> Result<(), Error> {
             self.terminate_tx.send(Ok(()))?;
             self.driver_hdl.await.unwrap()
         }
@@ -917,7 +948,7 @@ mod test {
     struct TestServerHandlerDriver {
         conn: WebSocketStream<TokioAdapter<TcpStream>>,
         event_rx: ChannelRx<Event>,
-        terminate_rx: ChannelRx<Result<()>>,
+        terminate_rx: ChannelRx<Result<(), Error>>,
         // A mapping of subscription queries to subscription IDs for this
         // connection.
         subscriptions: HashMap<String, String>,
@@ -927,7 +958,7 @@ mod test {
         fn new(
             conn: WebSocketStream<TokioAdapter<TcpStream>>,
             event_rx: ChannelRx<Event>,
-            terminate_rx: ChannelRx<Result<()>>,
+            terminate_rx: ChannelRx<Result<(), Error>>,
         ) -> Self {
             Self {
                 conn,
@@ -937,7 +968,7 @@ mod test {
             }
         }
 
-        async fn run(mut self) -> Result<()> {
+        async fn run(mut self) -> Result<(), Error> {
             loop {
                 tokio::select! {
                     Some(msg) = self.conn.next() => {
@@ -962,7 +993,7 @@ mod test {
             let _ = self.send(Id::Str(subs_id), ev).await;
         }
 
-        async fn handle_incoming_msg(&mut self, msg: Message) -> Option<Result<()>> {
+        async fn handle_incoming_msg(&mut self, msg: Message) -> Option<Result<(), Error>> {
             match msg {
                 Message::Text(s) => self.handle_incoming_text_msg(s).await,
                 Message::Ping(v) => {
@@ -977,7 +1008,7 @@ mod test {
             }
         }
 
-        async fn handle_incoming_text_msg(&mut self, msg: String) -> Option<Result<()>> {
+        async fn handle_incoming_text_msg(&mut self, msg: String) -> Option<Result<(), Error>> {
             match serde_json::from_str::<serde_json::Value>(&msg) {
                 Ok(json_msg) => {
                     if let Some(json_method) = json_msg.get("method") {
