@@ -3,7 +3,15 @@
 use std::str::FromStr;
 use std::{fs, path::PathBuf};
 use subtle_encoding::{base64, hex};
-use tendermint_rpc::{endpoint, error::Code, request::Wrapper as RequestWrapper, Order, Response};
+use tendermint::abci::transaction::Hash;
+use tendermint::evidence::Duration;
+use tendermint::public_key;
+use tendermint_rpc::{
+    endpoint,
+    error::{Error, ErrorDetail},
+    request::Wrapper as RequestWrapper,
+    Code, Order, Response,
+};
 use walkdir::WalkDir;
 
 const CHAIN_ID: &str = "dockerchain";
@@ -131,6 +139,14 @@ fn outgoing_fixtures() {
                         .unwrap();
                 assert_eq!(wrapped.params().height.unwrap().value(), 10);
             }
+            "consensus_params" => {
+                let wrapped = serde_json::from_str::<
+                    RequestWrapper<endpoint::consensus_params::Request>,
+                >(&content)
+                .unwrap();
+                let height = wrapped.params().height.unwrap();
+                assert_eq!(u64::from(height), 10u64);
+            }
             "consensus_state" => assert!(serde_json::from_str::<
                 RequestWrapper<endpoint::consensus_state::Request>,
             >(&content)
@@ -224,6 +240,19 @@ fn outgoing_fixtures() {
                     base64::decode("dHg1PXZhbHVl").unwrap()
                 );
             }
+            "tx" => {
+                let wrapped =
+                    serde_json::from_str::<RequestWrapper<endpoint::tx::Request>>(&content)
+                        .unwrap();
+                assert_eq!(
+                    wrapped.params().hash,
+                    Hash::new([
+                        214, 63, 156, 35, 121, 30, 97, 4, 16, 181, 118, 216, 194, 123, 181, 174,
+                        172, 147, 204, 26, 88, 82, 36, 40, 167, 179, 42, 18, 118, 8, 88, 96
+                    ])
+                );
+                assert!(!wrapped.params().prove);
+            }
             "tx_search_no_prove" => {
                 let wrapped =
                     serde_json::from_str::<RequestWrapper<endpoint::tx_search::Request>>(&content)
@@ -306,15 +335,20 @@ fn incoming_fixtures() {
                 assert!(result.response.value.is_empty());
             }
             "block_at_height_0" => {
-                let error = endpoint::block::Response::from_string(&content)
-                    .err()
-                    .unwrap();
-                assert_eq!(error.code(), Code::InternalError);
-                assert_eq!(error.message(), "Internal error");
-                assert_eq!(
-                    error.data(),
-                    Some("height must be greater than 0, but got 0")
-                );
+                let res = endpoint::block::Response::from_string(&content);
+
+                match res {
+                    Err(Error(ErrorDetail::Response(e), _)) => {
+                        let response = e.source;
+                        assert_eq!(response.code(), Code::InternalError);
+                        assert_eq!(response.message(), "Internal error");
+                        assert_eq!(
+                            response.data(),
+                            Some("height must be greater than 0, but got 0")
+                        );
+                    }
+                    _ => panic!("expected Response error"),
+                }
             }
             "block_at_height_1" => {
                 let result = endpoint::block::Response::from_string(content).unwrap();
@@ -615,6 +649,26 @@ fn incoming_fixtures() {
                     tendermint::block::header::Version { block: 11, app: 1 }
                 );
             }
+            "consensus_params" => {
+                let result = endpoint::consensus_params::Response::from_string(content).unwrap();
+                assert_eq!(u64::from(result.block_height), 10_u64);
+                assert_eq!(result.consensus_params.block.max_bytes, 22020096_u64);
+                assert_eq!(result.consensus_params.block.max_gas, -1_i64);
+                assert_eq!(result.consensus_params.block.time_iota_ms, 500_i64);
+                assert_eq!(
+                    result.consensus_params.evidence.max_age_duration,
+                    Duration(std::time::Duration::from_nanos(172800000000000_u64))
+                );
+                assert_eq!(
+                    result.consensus_params.evidence.max_age_num_blocks,
+                    100000_u64
+                );
+                assert_eq!(result.consensus_params.evidence.max_bytes, 1048576_i64);
+                assert_eq!(
+                    result.consensus_params.validator.pub_key_types,
+                    vec![public_key::Algorithm::Ed25519]
+                );
+            }
             "consensus_state" => {
                 assert!(endpoint::consensus_state::Response::from_string(content).is_ok());
             }
@@ -724,23 +778,26 @@ fn incoming_fixtures() {
                 assert_eq!(result.validator_info.power.value(), 10);
             }
             "subscribe_malformed" => {
-                let result = endpoint::subscribe::Response::from_string(content)
-                    .err()
-                    .unwrap();
-                assert_eq!(result.code(), Code::InternalError);
-                assert_eq!(result.message(), "Internal error");
-                assert_eq!(result.data().unwrap(),"failed to parse query: \nparse error near PegText (line 1 symbol 2 - line 1 symbol 11):\n\"malformed\"\n");
+                let result = endpoint::subscribe::Response::from_string(content);
+
+                match result {
+                    Err(Error(ErrorDetail::Response(e), _)) => {
+                        let response = e.source;
+
+                        assert_eq!(response.code(), Code::InternalError);
+                        assert_eq!(response.message(), "Internal error");
+                        assert_eq!(response.data().unwrap(),"failed to parse query: \nparse error near PegText (line 1 symbol 2 - line 1 symbol 11):\n\"malformed\"\n");
+                    }
+                    _ => panic!("expected Response error"),
+                }
             }
             "subscribe_newblock" => {
-                let result = endpoint::subscribe::Response::from_string(content)
-                    .err()
-                    .unwrap();
-                assert_eq!(result.code(), Code::ParseError);
-                assert_eq!(result.message(), "Parse error. Invalid JSON");
-                assert_eq!(
-                    result.data().unwrap(),
-                    "missing field `jsonrpc` at line 1 column 2"
-                );
+                let result = endpoint::subscribe::Response::from_string(content);
+
+                match result {
+                    Err(Error(ErrorDetail::Serde(_), _)) => {}
+                    _ => panic!("expected Serde parse error, instead got {:?}", result),
+                }
             }
             "subscribe_newblock_0" => {
                 let result = tendermint_rpc::event::Event::from_string(content).unwrap();
@@ -1300,6 +1357,17 @@ fn incoming_fixtures() {
                 );
                 assert!(result.log.value().is_empty());
             }
+            "tx" => {
+                let result = endpoint::tx::Response::from_string(content).unwrap();
+                assert_eq!(
+                    result.hash,
+                    Hash::new([
+                        214, 63, 156, 35, 121, 30, 97, 4, 16, 181, 118, 216, 194, 123, 181, 174,
+                        172, 147, 204, 26, 88, 82, 36, 40, 167, 179, 42, 18, 118, 8, 88, 96
+                    ])
+                );
+                assert_eq!(u64::from(result.height), 12u64);
+            }
             "tx_search_no_prove" => {
                 let result = endpoint::tx_search::Response::from_string(content).unwrap();
                 assert_eq!(result.total_count as usize, result.txs.len());
